/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_data.h"
#include "imaging_fdata.h"
#include "minrtsFftUtil.h"
#ifdef HP
#define finite isfinite
#endif

#define XSMOOTH 11

/**********************************************************
 * Usage:
 *    oirMakeMask "AFiles" "BFiles" outFile VIS|TRACK [wFudge]
 *
 *    AFiles and BFiles are  blank-separated lists of input files
************************************************************** */
void Usage() {
   printf("%s\n", "Usage:");
   printf("%s\n\n", "oirMakeMask AShutterFiles BShutterFiles outFile VIS|TRACK [wFudge]");
   printf("%s\n", "AShutterFiles = Filenames of exposure with Shutter A(only) open.");
   printf("%s\n", "BShutterFiles = Filenames of exposure with Shutter B(only) open.");
   printf("%s\n", "outFile       = name of output file.");
   printf("%s\n", "VIS|TRACK: specify whether a Visibility estimate or Tracking mask wanted.");
   printf("%s\n", "[wFudge]: optional fudge factor; width of mask will be increased by this factor,");
   printf("%s\n", "   if given, else default = 1.0");
   return;
}

void setBadPix(int nPix, float *testFrame, int *badPix) {
/***************************************************************
 *  from the statistics in testFrame try to find bad
    pixels and set them in badPix to 0
************************************************************** */
   /* dummy */
   int i;
   for (i=0;i<nPix;i++) badPix[i] = 1;
   return;
   }
/***************************************************************
 *  A really stupid gaussian fitting routine (cant even spell 
    gauss right !)
************************************************************** */
   void dumGoose(int nData, float *x, float *y, float *weight, 
      float* fit, int niter) {
   float x0, p0;
   float dfit[3], tfit[3];
   float chi2[5];
   float xx[5];
   float t;

   int iter,i,j,iData;
   int iPar; /* which parameter to vary */
/*
      find initial position and magnitude of peak
      put into initial fit vector; assume initial
      width = 1.
*/
   x0 = minrtsQPk(nData, x, y, &p0,0);
   fit[0] = x0;
   fit[1] = p0;
   fit[2] = 1.;

/*   dfit[0] = fit[0] * .3; */
   dfit[0] = 0.5 * (x[nData-1]-x[nData])/nData;
   dfit[1] = fit[1] * .1;
   dfit[2] = fit[2] * .1;

/*
     iterate a bunch of times;  for each iteration
     try changing each of the variables and move
     to minimum of chi2
*/
   for (iter=0; iter<niter;iter++) {
      for(iPar=0;iPar<3;iPar++){
         for (i=0;i<5;i++)    xx[i] = fit[iPar] + (i-2)*dfit[iPar];
         for (i=0;i<5;i++) {
             for (j=0;j<3;j++) tfit[j] = fit[j];
             tfit[iPar] = xx[i];
             chi2[i] = 0.;
             for (iData=0;iData<nData;iData++) {
                t = (tfit[0]-x[iData])*(tfit[0]-x[iData])/(tfit[2]*tfit[2]);
                if (t < 20) t = tfit[1]*exp(-0.5*t); else t = 0.;
                chi2[i] = chi2[i] + (y[iData] - t)*(y[iData] - t);
             }  /* iData loop */
         }  /* i=-2,-1,0,1,2 loop */
         fit[iPar] = minrtsQPk(5, xx, chi2, &p0,1);
      } /* iPar */
   } /* iter */
   return;
} /* end of dumGoose */

/***************************************************************
 * Split a blank-separated set of file names into separate character strings
 * INPUTS:
 *    char *string     blank-separated list
 * OUTPUTS:
 *    int    nString      number of strings found
 * RETURNS
 *    char **   array of output file names; these have been malloced and
 *              should be freed by calling program killFileList
 **************************************************************** */
char** strSplit (char *string, int *nString) {
   char *blank = " ";
   char *oldString;
   int   nS, iS ;
   char *tok;
   char *stringCopy;
   char **output;

   *nString = 0;
/*            copy input string                */
   stringCopy = (char *)malloc(sizeof(char) * (1+strlen(string)));
   strcpy (stringCopy, string);
   oldString = stringCopy;
/*          find pointer to first blank delimited string   */
   tok = strtok(oldString, blank);
   if (tok == NULL) return NULL;
   oldString = NULL;
/*          get pointers to the rest of the sub strings */
   for (nS = 0; tok != NULL; nS++) {
      tok = strtok(oldString, blank);
   }
   *nString = nS;
/*          malloc output arrays and copy strings */
   output = (char **)malloc(nS * sizeof(char *));
   strcpy (stringCopy, string);
   oldString = stringCopy;
   blank = " ";
   for (iS=0; iS<nS; iS++) {
      tok = strtok(oldString, blank);
      oldString = NULL;
      output[iS] = (char *)malloc(sizeof(char) * (1+strlen(tok)));
      strcpy(output[iS], tok);
   }
   free(oldString);
   return output;
} 

/***************************************************************
 * kill space associated with array of file names
 * INPUTS:
 *    char **string     array of file names
 *    int    nFiles     number of file names
 **************************************************************** */
void killFileList(char **fileList, int nFiles) {
   int iFile;
   for (iFile=0; iFile<nFiles;iFile++) {free(fileList[iFile]);}
   free(fileList);
   return;
} /* end of killFileList */

/***************************************************************
 * read through all raw files associated with one shutter setting
 * separate target and sky frames and accumulate them in
 * given arrays; divide by number of frames
 * INPUTS:
 *    char **string     array of file names
 *    int    nFiles     number of file names
 *    int    chan1      0-based number of 1st interferometric channel
 *    int    chan2      0-based number of 2nd interferometric channel
 *    int    nDrop      drop this many points after change of chop phase
 *    int    npix1      no. pixels in array in 1st channel
 *    int    npix2      no. pixels in array in 2nd channel
 *    float* target1    array to receive target1 data
 *    float* target2    array to receive target2 data
 *    float* sky1       array to receive target1 data
 *    float* sky2       array to receive target2 data
 *    int  * nTarget    number of target frames
 *    int  * nSky       number of sky frames
 **************************************************************** */

void unChopData(char **fileList, int nFiles, int nChan, 
   int nDrop, long npix1, long npix2, float **target, float **sky, 
      int *nTarget, int *nSky, int *ierr) {

   fitsfile *inFile;
   imaging_data_table_def *idata;
   imaging_data_row *inData;
   int iFile;

   long iRow, nRow;
   int i;
   int TN, SN;
   long iPix;
   int iChan;

   char Sky = 'S';
   char Target = 'T';
   char targ, oldTarg;
   int sinceChange;

   *ierr = 0;
   TN = 0;
   SN = 0;
   oldTarg = ' ';

/*
      loop over files, read rows, read target types, add data
      into appropriate accumulators 
*/

   for (iFile =0; iFile <nFiles; iFile++) {
/*
      open file, open table, create table row
*/
      i = fits_open_file(&inFile, fileList[iFile], READONLY, ierr);
      if (*ierr !=0) {
         printf("error %i opening file %s\n", *ierr, fileList[iFile]);
         printf("%s \n",strerror(errno));
         return;
      }
      idata = open_imaging_data_table(inFile, ierr);
      if (*ierr !=0) {
         printf("error %i opening imaging data in file %s\n", 
            *ierr, fileList[iFile]);
         return;
      }
      inData = make_imaging_data_row(idata);
      i = fits_get_num_rows (inFile,  &nRow, ierr);
      /* suppress autoscaling to avoid overflow */
      for (i=0; i<idata->nregion; i++)
         fits_set_tscale(inFile, 1+idata->data_col[i], 1.0, 0.0, ierr);

/*
         read lines, check target type and decide whether to accumulate
*/
      for (iRow =1; iRow<=nRow; iRow++) {
         i = read_imaging_data_table(inFile, idata, inData, iRow, ierr);
         targ = inData->targtype[0];
         if (targ != oldTarg ) {
            sinceChange = 0;
            oldTarg     = targ;
         } else {
            ++sinceChange;
         }
/*
            if sinceChange > nDrop decide what type and accumulate data
*/
         if (sinceChange > nDrop && (targ == Target)) {
            for (iChan=0;iChan<nChan;iChan++) {
               for (iPix=0;iPix<npix1;iPix++) target[iChan][iPix] =
               target[iChan][iPix] + 32768. + (short)inData->data[iChan][iPix];
               for (iPix=0;iPix<npix2;iPix++) target[iChan][iPix] =
               target[iChan][iPix] + 32768. + (short)inData->data[iChan][iPix];
            }
            ++TN;
         } else if (sinceChange > nDrop && (targ == Sky)) {
            for (iChan=0;iChan<nChan;iChan++) {
               for (iPix=0;iPix<npix1;iPix++) sky[iChan][iPix] =
                  sky[iChan][iPix] + 32768. + (short)inData->data[iChan][iPix];
               for (iPix=0;iPix<npix2;iPix++) sky[iChan][iPix] =
                  sky[iChan][iPix] + 32768. + (short)inData->data[iChan][iPix];
            }
            ++SN;
         }
      } /* read rows in one file */
      
/*
         close file and table, destroy row
*/
      i= close_imaging_data_table(inFile, idata, ierr);
      kill_imaging_data_row(inData);
   } /* read all files in list */
   *nTarget = TN;
   *nSky    = SN;
   return;
} /* end of unChopData */

void normalizeExposure (float *data, long nPix, int nFrames, float dit) {
   float norm;
   long iPix;

   if (nFrames <= 0) return;
   norm = 1./(nFrames * dit);
   for (iPix =0; iPix<nPix; iPix++) data[iPix] = data[iPix]*norm;
   return;
}

void deSky (float *data, int nx, int ny, int y1, int y2) {
   int ix, iy;
   float dy, x1, x2;
   float *topSky, *bottomSky;
   float *yLine;
/*
      arrays to hold average of a few rows at top and bottom
*/
   topSky = (float *)malloc(nx * sizeof(float));
   memset(topSky, 0, nx*sizeof(float));
   bottomSky = (float *)malloc(nx * sizeof(float));
   memset(bottomSky, 0, nx*sizeof(float));

/*
      average a few rows
*/
   for (iy =-1; iy<=1; iy++) for (ix=0;ix<nx;ix++) {
      topSky[ix]    = topSky[ix] + data[ix + (iy+y2)*(long)nx];
      bottomSky[ix] = bottomSky[ix] + data[ix + (iy+y1)*(long)nx];
   }
   for (ix=0;ix<nx;ix++) {
      topSky[ix]    = topSky[ix]/3.;
      bottomSky[ix] = bottomSky[ix]/3.;
   }
   dy = y2 - y1;

/*
      linear fit and subtraction
*/
   for (iy=0; iy<ny; iy++) {
      yLine = data+iy*(long)nx;
      x1 = ((float)iy - y1)/dy;
      x2 = ((float)iy - y2)/dy;
      for (ix=0; ix<nx;ix++) yLine[ix] = yLine[ix] - topSky[ix]*x1 + bottomSky[ix]*x2;
   }

/*
      free arrays
*/
   free(topSky);
   free(bottomSky);
}

void trackMask (int nx, int ny, int *badPix, float *AData, float *BData, float *mask,
   float wFudge) {

   float *aFit[3], *bFit[3], *xx, *yy, *ww0, *ww;
   float x;
   int i, ix, iy;
   float gFit[3];

   for (i=0;i<3;i++) {
      aFit[i] = (float *)malloc(nx * sizeof(float));
      bFit[i] = (float *)malloc(nx * sizeof(float));
   }
   xx = (float *)malloc(ny * sizeof(float)); /* x-coordinate */
   yy = (float *)malloc(ny * sizeof(float)); /* intensities */
   ww = (float *)malloc(ny * sizeof(float)); /* weights */
   ww0 = (float *)malloc(ny * sizeof(float)); /*default weights */

   for (iy=0;iy<ny;iy++) xx[iy] = (float)iy + 1;
   for (iy=0;iy<ny;iy++) ww0[iy] = 1.;
   for (i=0;i<4;i++) {
      ww0[i] = 0.;
      ww0[ny-i-1] = 0.;
   }

   for (ix = 0; ix<nx; ix++) {
/*
       copy default weights * badPix -> weights
*/
      for (iy=0;iy<ny;iy++)  ww[i] = ww0[i]*badPix[ix*nx + iy];

/*
       copy A data -> yy and fit gaussian; then B data
*/
      for (iy=0;iy<ny;iy++)  yy[iy] = AData[iy*nx + ix];
      dumGoose(ny, xx, yy, ww, gFit, 10);
      for (i=0;i<3;i++) aFit[i][ix] = gFit[i];
      for (iy=0;iy<ny;iy++)  yy[i] = BData[iy*nx + ix];
      dumGoose(ny, xx, yy, ww, gFit, 10);
      for (i=0;i<3;i++) bFit[i][ix] = gFit[i];
   }  /* ix loop */
/*
      edit fit parameters to eliminate duds and smooth noise
*/
   for (ix=0; ix<nx; ix++) {
      if (!finite(aFit[0][ix])) aFit[0][ix] = .5*ny;
      if (!finite(aFit[1][ix])) {
         aFit[1][ix] = 0.;
         bFit[1][ix] = 0.;
      }
      if (!finite(aFit[2][ix])) aFit[2][ix] = 1.;
      if (!finite(bFit[0][ix])) bFit[0][ix] = .5*ny;
      if (!finite(bFit[1][ix])) {
         aFit[1][ix] = 0.;
         bFit[1][ix] = 0.;
      }
      if (!finite(bFit[2][ix])) bFit[2][ix] = 1.;
   }
   x = 0;
   for (ix=0; ix<nx; ix++) if (aFit[1][ix] > x) x=aFit[1][ix];
   if (x > 0) for (ix=0; ix<nx; ix++) aFit[1][ix]  = aFit[1][ix]/x;
   x = 0;
   for (ix=0; ix<nx; ix++) if (bFit[1][ix] > x) x=bFit[1][ix];
   if (x > 0) for (ix=0; ix<nx; ix++) bFit[1][ix]  = bFit[1][ix]/x;
/*
      average A- and B- fits and create mask
*/
   for (ix =0; ix<nx; ix++) {
      for (i=0; i<3; i++) aFit[i][ix] = 0.5*(aFit[i][ix]+bFit[i][ix]);
      aFit[2][ix] = wFudge * aFit[2][ix];
      for (iy=0; iy<ny; iy++){
         x = 0.5*((xx[iy]-aFit[0][ix])*(xx[iy]-aFit[0][ix])/(aFit[2][ix]*aFit[2][ix]));
         if (x > 20.) x = 20.;
         mask[iy*nx+ix] = aFit[1][ix]*exp(-x)*badPix[iy*nx+ix];
      }
   }
   

/*      clean up   */

   for (i=0;i<3;i++) {
      free(aFit[i]);
      free(bFit[i]);
   }
   free(xx);
   free(yy);
   free(ww0);
   free(ww);

   return;
} /* end of trackMask */

void visMask(int AB, int nChan, int nx, int ny, int* badPix[], float *data[], 
   float* mask[], float wFudge) {

   float *xx, *yy, *ww0, *ww;
   float *xs, *ys;
   float x, y;
   int i, j, ix, iy, iChan, ns, ns2;
   float gFit[3];

   xx = (float *)malloc(ny * sizeof(float)); /* x-coordinate */
   yy = (float *)malloc(ny * sizeof(float)); /* intensities */
   ww = (float *)malloc(ny * sizeof(float)); /* weights */
   ww0 = (float *)malloc(ny * sizeof(float)); /*default weights */
   xs = (float *)malloc(nx * sizeof(float)); /* smoothing array */
   ys = (float *)malloc(nx * sizeof(float)); /* smoothing array */

   for (iy=0;iy<ny;iy++) xx[iy] = (float)iy + 1;
   for (iy=0;iy<ny;iy++) ww0[iy] = 1.;
   for (i=0;i<4;i++) {
      ww0[i] = 0.;
      ww0[ny-i-1] = 0.;
   }

   for (iChan =0; iChan<nChan; iChan++) {
      if (((AB ==0 && (iChan !=3))) || ((AB==1) && (iChan !=0) || (nChan != 4))) 
      for (ix = 0; ix<nx; ix++) {
/*
       copy default weights * badPix -> weights
*/
         for (iy=0;iy<ny;iy++)  ww[i] = ww0[i]*badPix[iChan][iy*nx + ix];

/*
       copy data -> yy and fit gaussian
*/
         for (iy=0;iy<ny;iy++)  yy[iy] = data[iChan][iy*nx + ix];
         dumGoose(ny, xx, yy, ww, gFit, 10);
         for (i=0;i<3;i++) mask[iChan][i*nx + ix] = gFit[i];
      }  /* ix loop */
/*
      edit fit parameters to eliminate duds and smooth noise
      set normalization to 1.0 and apply wFudge
*/
      for (ix=0; ix<nx; ix++) {
         if (!finite(mask[iChan][ix])) mask[iChan][ix] = 0.5*ny;
         if (!finite(mask[iChan][nx+ix])) mask[iChan][nx+ix] = 0.; 
         if (!finite(mask[iChan][2*nx + ix])) mask[iChan][2*nx + ix] = 1.;
            else mask[iChan][2*nx+ix] = wFudge * mask[iChan][2*nx+ix];
         if (mask[iChan][2*nx + ix]<1. ||(mask[iChan][2*nx + ix]>5)) {
            mask[iChan][2*nx + ix] = wFudge;
            mask[iChan][ix] = 0.5*ny;
         }
      }
      x = 0;
      for (ix=0; ix<nx; ix++) if (mask[iChan][nx+ix] > x) x=mask[iChan][nx+ix];
      x = .03*x; /* drop columns below 3% of peak */
      for (ix=0; ix<nx; ix++) if (mask[iChan][nx+ix] > x) mask[iChan][nx+ix]=1.;
         else mask[iChan][nx+ix] = 0.;
/*
           box car smooth
*/
      ns2 = XSMOOTH/2;
      for (ix=0; ix<nx; ix++) {
         x  = 0.;
         y  = 0.;
         ns = 0;
 
         for (j=ix-ns2; j<=ix+ns2; j++) if (j>=0 && (j<nx)) 
            if(mask[iChan][nx+j]>0.) {
               x = x + mask[iChan][j];
               y = y + mask[iChan][2*nx+j];
               ns = ns + 1;
            }

         if (ns > 0) {
            xs[ix] = x/ns;
            ys[ix] = y/ns;
         } else {
            xs[ix] = mask[iChan][ix];
            ys[ix] = mask[iChan][2*nx+ix];
         }
      }
      for (ix=0; ix<nx; ix++) {
            mask[iChan][ix] = xs[ix];
            mask[iChan][2*nx+ix] = ys[ix];
      }

   } /* iChan loop */
   free(xx);
   free(yy);
   free(xs);
   free(ys);
   free(ww0);
   free(ww);

   return;
}


int main (int argc, char** argv) {

   char opt1[12];
   char shutter[12];
   float dit;

   char track; 
   char **fileList;
   int nFiles;
   int iShutter;
   int i,ierr;
   int nx, ny;
   long npix;
   long iPix;
   int  ix, iy;
   float xx, mm;
   long  ii;
   int iKey;
   int one = 1;

   int yTop = 34;
   int yBottom = 12;
   int chan1, chan2;

   char outFileName[130];
   char *bang="!";

   fitsfile *inFile;
   fitsfile *outFile = NULL;

   imaging_fdata_table_def *odata;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_row *outData;
   int outLine = 0;

   char *Ashutter = "AOPEN";
   char *Bshutter = "BOPEN";
   int  nShutter;
   char *highSens = "HIGH_SENS";
   int  nChan, iChan;

   float *ATarget[4], *BTarget[4], *ASky[4], *BSky[4];
   int *badPix[4]; 
   int    ATargetN, ASkyN, BTargetN, BSkyN;
   float *sumX, sumY;

   float *mask[4];

   int nDrop;
   float wFudge;

   ierr = 0;
   nDrop = 3;
/*
      check input
*/
   if (argc < 5 || (argc > 6)) {
      Usage();
      return 1;
   }
/*
      see if TRACK or VIS mask is desired
*/
      if ((*argv[4]) == 'T') track =1;
      else if ((*argv[4])=='V') track = 0;
      else {
         Usage();
         return 1;
      }
      if (track) printf("making TRACKING mask\n");
         else printf("making VISIBILITY mask\n");
/*
      see if wFudge is specified externally
*/
   if (argc > 5) sscanf(argv[5], "%f", &wFudge);
      else wFudge = 1.;
   printf("wFudge = %f\n", wFudge);

/*        
       break up file list and print what you find 
       then read and unchop files
*/
   for (iShutter = 1;iShutter<=2; iShutter++) {
      fileList = strSplit(argv[iShutter], &nFiles);

/*
      open first file and get a lot of parameters for iShutter=1
      also open output file, initialize it and copy header info.
*/
      i = fits_open_file(&inFile, fileList[0], READONLY, &ierr) ;
      if (ierr != 0) {
         printf("cfitsio error %i opening file %s\n", ierr, fileList[0]);
         printf("%s \n",strerror(errno));
         return 1;
      }
      ierr = 0;
      i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS SHUT NAME", 
         shutter, 0, &ierr);
      printf("shutter = %s\n", shutter);
      nShutter = 0;
      if (strncmp(shutter, Ashutter, 5)==0) nShutter=1;
      if (strncmp(shutter, Bshutter, 5)==0) nShutter=2; 

      if (iShutter == 1) {
         i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO DET DIT", &dit, 
            0, &ierr);
         printf("dit = %f\n", dit);
         ierr = 0;
         i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS OPT1 NAME", opt1,             0, &ierr);
         printf("beam combiner = %s\n", opt1);

         if (strncmp(opt1, highSens, 9) == 0) { 
            nChan = 2; 
            chan1 = 0;
            chan2 = 1;
         } else {
            nChan = 4;
            chan1 = 1;
            chan2 = 2;
         }
/*
      prepare output file
*/
         strcpy(outFileName, bang);
         strcat(outFileName, argv[3]);

         fits_create_file (&outFile, outFileName, &ierr);
         if (ierr !=0) {
            fprintf (stderr, "error opening fits output %s %i\n",argv[3], ierr);
            printf("%s \n",strerror(errno));
            return 1;
         }
/*
           copy primary header from input to output
*/
         i = fits_copy_header(inFile, outFile, &ierr);

/*
      move to detector section
*/
         DetectorDef = read_imaging_detector_table(inFile, &ierr);
         if (ierr !=0) {
            printf("error %i opening detector table in %s\n", 
               ierr, fileList[0]);
            return 1;
         }
         nx = DetectorDef->naxis[0][0];
         ny = DetectorDef->naxis[0][1];

/*
      allocate accumulator arrays
*/
  
         npix = ny * (long)nx;
         for (iChan=0; iChan<nChan;iChan++) {
            ATarget[iChan] = (float *)malloc(npix * sizeof(float));
            memset (ATarget[iChan], 0, npix*sizeof(float));
            BTarget[iChan] = (float *)malloc(npix * sizeof(float));
            memset (BTarget[iChan], 0, npix*sizeof(float));
            ASky[iChan] = (float *)malloc(npix * sizeof(float));
            memset (ASky[iChan], 0, npix*sizeof(float));
            BSky[iChan] = (float *)malloc(npix * sizeof(float));
            memset (BSky[iChan], 0, npix*sizeof(float));
            badPix[iChan] = (int *)malloc(npix * sizeof(int));
            memset (BSky[iChan], 1, npix);
            if (track) {
               mask[iChan] = (float *)malloc(npix * sizeof(float));
               memset (mask[iChan], 0, npix*sizeof(float));
            } else {
               mask[iChan] = (float *)malloc(3*nx * sizeof(float));
               memset (mask[iChan], 0, 3*nx*sizeof(float));
            }
         }
/*
           if we are outputing VIS tables, modify images sizes in
           DetectorDef
*/
         if (! track) {
            for (iChan=0;iChan<nChan;iChan++) DetectorDef->naxis[iChan][1] = 3;
            i = fits_write_key(outFile, TSTRING, "MASKTYPE", "GAUSS/VISIBILITY"," ",&ierr);
         } else {
            i = fits_write_key(outFile, TSTRING, "MASKTYPE", "IMAGE/TRACKING"," ",&ierr);
         }
         odata = make_imaging_fdata_table_def (outFile, DetectorDef, 1, &ierr);
         if (ierr !=0) {
            fprintf (stderr, "error creating data definition\n");
            return 1;
         }
         outData = make_imaging_fdata_row(odata);

      }  /* iShutter = 1 */
      i = fits_close_file(inFile, &ierr);
/*
          now process the real data
*/
      if (nShutter == 1) {
         printf("Found %i A-photometry Files\n", nFiles);
         for (i=0; i<nFiles; i++) printf("%s \n", fileList[i]);
         unChopData(fileList, nFiles, nChan, nDrop,
            npix, npix, ATarget, ASky, &ATargetN, &ASkyN,
            &ierr);
      }
      if (nShutter == 2) {
         printf("Found %i B-photometry Files\n", nFiles);
         for (i=0; i<nFiles; i++) printf("%s \n", fileList[i]);
         unChopData(fileList, nFiles, nChan, nDrop,
            npix, npix, BTarget, BSky, &BTargetN, &BSkyN,
            &ierr);
      }
      killFileList(fileList, nFiles);
   }   /* loop over shutters */
/*
       normalize data by dividing by exposure time
*/
   for (iChan =0; iChan<nChan; iChan++) {
      normalizeExposure(ATarget[iChan], npix, ATargetN, dit);
      normalizeExposure(BTarget[iChan], npix, BTargetN, dit);
      normalizeExposure(ASky[iChan], npix, ASkyN, dit);
      normalizeExposure(BSky[iChan], npix, BSkyN, dit);
   }
/*
      set badpix masks from sky images
*/
   for (iChan =0; iChan<nChan; iChan++) setBadPix(npix, ASky[iChan], badPix[iChan]);

/*  for now write out raw A values for debug 
   for (iChan=0; iChan<nChan; iChan++) 
      for (ix=0; ix<npix; ix++) outData->data[iChan][ix] = ATarget[iChan][ix];
      i = write_imaging_fdata_table(outFile, odata, outData, ++outLine, &ierr);
*/

/*
      subtract sky values
*/
   for (iChan =0; iChan<nChan; iChan++) {
      for (iPix = 0; iPix< npix; iPix++) {
         ATarget[iChan][iPix] = ATarget[iChan][iPix] - ASky[iChan][iPix];
         BTarget[iChan][iPix] = BTarget[iChan][iPix] - BSky[iChan][iPix];
      }
   }
/*
      deSky
      y coordinates of sky are cludge for prism; need better algorithm to find background
*/
   for (iChan =0; iChan<nChan; iChan++) {
      deSky(ATarget[iChan], nx, ny, yBottom, yTop); 
      deSky(BTarget[iChan], nx, ny, yBottom, yTop);
   }

/*
         if debugging write output data

   for (iChan=0; iChan<nChan; iChan++) 
      for (ix=0; ix<npix; ix++) outData->data[iChan][ix] = ATarget[iChan][ix];
      i = write_imaging_fdata_table(outFile, odata, outData, ++outLine, &ierr);
   for (iChan=0; iChan<nChan; iChan++) 
      for (ix=0; ix<npix; ix++) outData->data[iChan][ix] = BTarget[iChan][ix];
      i = write_imaging_fdata_table(outFile, odata, outData, ++outLine, &ierr);

*/

/*
       decide whether to generate a fringe tracking mask (only interferometric
       channels) or a visibility estimation mask
*/

   if (track) {
      trackMask(nx, ny, badPix[chan1], ATarget[chan1], BTarget[chan1], mask[chan1],wFudge);
      trackMask(nx, ny, badPix[chan2], ATarget[chan2], BTarget[chan2], mask[chan2],wFudge);
      for (iChan=0; iChan<nChan; iChan++) 
         for (ix=0; ix<npix; ix++) outData->data[iChan][ix] = mask[iChan][ix];
         i = write_imaging_fdata_table(outFile, odata, outData, ++outLine, &ierr);

   } else {
      visMask(0, nChan, nx, ny, badPix, ATarget, mask, wFudge);
      for (iChan=0; iChan<nChan; iChan++) 
         for (ix=0; ix<(3*nx); ix++) outData->data[iChan][ix] = mask[iChan][ix];
         i = write_imaging_fdata_table(outFile, odata, outData, ++outLine, &ierr);
      visMask(1, nChan, nx, ny, badPix, BTarget, mask, wFudge);
      for (iChan=0; iChan<nChan; iChan++) 
         for (ix=0; ix<(3*nx); ix++) outData->data[iChan][ix] = mask[iChan][ix];
         i = write_imaging_fdata_table(outFile, odata, outData, ++outLine, &ierr);
   }
   i = close_imaging_fdata_table(outFile, odata, &ierr);

   kill_imaging_fdata_row(outData);

   for (iChan=0;iChan<nChan;iChan++) {
      free(ATarget[iChan]);
      free(BTarget[iChan]);
      free(ASky[iChan]);
      free(BSky[iChan]);
      free(badPix[iChan]);
      free(mask[iChan]);
   }
   return 0;
}
